/*
 *  auditd.c
 *  Linux Basic Security Module 
 *  Copyright (c) 2000 The Regents of the University of California
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright notice,
 *  this list of conditions and the following disclaimer.
 *
 *
 *  2. Redistributions in binary form must reproduce the above copyright notice,
 *  this list of conditions and the following disclaimer in the documentation
 *  and/or other materials provided with the distribution.
 *
 *  3. The name of the author may not be used to endorse or promote products
 *  derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND 
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
 *  SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 *  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 *  GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 *  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <ctype.h>
#include <errno.h>		/* perror() */
#include <fcntl.h>
#include <signal.h>
#include <stdint.h>		/* uint32_t	*/
#include <stdio.h>
#include <time.h>		/* time_t */
#include <unistd.h>		/* pause(), unlink() */

#include <sys/mman.h>	/* mmap() */
#include <sys/stat.h>	/* file modes */
#include <sys/types.h>

#include "auditd.h"

pref* user_prefs;			/* Daemon prefs */
int process_id;				/* Daemon's process id */
int busy;					/* Flag, set if writing out to the outfile */
char* buffer1;				/* The primary data area */
char* buffer2;				/* The secondary data area */
const char device_name[] = "/dev/auditor";	/* the device name */
int device_fd;				/* File descriptor for /dev/auditor */
char trail_name[PATHLEN];	/* The audit trail file name */
int trail_fd;				/* The trail's corresponding file descriptor */
char trail_name_old[PATHLEN]; /* needed for renaming at the end */
char start_time[15];	/* saves the trail starting time */


/* Builds a pref struct with set defaults */
pref* new_prefs(void)
{
	pref* prefs;

	/* Allocate memory and fill it in */
	if((prefs = (pref*) malloc(sizeof(pref))) == NULL)
		error_sys("Couldn't allocate memory for preferences\n");
		
	prefs->default_flag = 0;
	prefs->auditing_pid = 0;
	prefs->auditing_gid = 0;
	prefs->auditing_uid = 0;
	prefs->custom_bufsize = 0;
	prefs->pid = 0;
	prefs->gid = 0;
	prefs->uid = 0;
	prefs->bufsize = BSM_DEFAULT_BUF;
	
	return prefs;
}

/* Result of calling with bad command-line flags */
void bad_prefs(char val)
{
	
	switch(val) {
	default: printf("unrecognized option\n");
	}
	
	printf("usage: auditd -bgpu num \n");
	exit(1);
}

/* process command-line arguments, set up prefs
	-b size Sets a buffer size in K (well, thousands of bytes)
	-g GID audit one group
	-p PID audit one pid
	-u UID audit one user
	
Yeah, unable to set more than one option for now
or to audit for specific event class(es)
*/   
void set_prefs(int argc, char* argv[], pref* user_prefs)
{
	int i, j;
	int opt_length;   /* Length of option string */
	
	/* if no arg, set defaults */
	if (argc == 1)
    	user_prefs->default_flag = 1;
    else
	{
		/* Make sure this at least could be a valid option */
		if ((argc != 3) || (argv[1][0] != '-'))
			bad_prefs(0);

		/* Now parse out the options */
		switch(argv[1][1])
		{
		case 'b':
		case 'B':
			/* If so, get the buffer size */
			user_prefs->custom_bufsize = 1;
			user_prefs->bufsize = strtoul(argv[2], (char**) NULL, 10);
			user_prefs->bufsize *= 1000;
			break;
		
		case 'g':
			user_prefs->auditing_gid = 1;
			user_prefs->gid = strtoul(argv[2], (char**) NULL, 10);
			break;

		case 'p':
		case 'P':
			user_prefs->auditing_pid = 1;
			user_prefs->pid = strtoul(argv[2], (char**) NULL, 10);
			break;

		case 'u':
		case 'U':
			user_prefs->auditing_uid = 1;
			user_prefs->uid = strtoul(argv[2], (char**) NULL, 10);
			break;

		default :
			bad_prefs(1);	  
		}
	}

}

/* Builds up a file name for the audit trail we're about to start */
/* opt = 0 for a new (not_terminated) filename, 1 to get starttime.endtime */
/* start_time and timesize are global */
void set_filename(int opt)
{
	time_t temp_time;
	struct tm *binary_time;
	const int timesize = 15; /* size of formatted tm struct plus a null */
	const char fmt[] = "%Y%m%d%H%M%S";
	const char period[] = ".";
	const char not_term[] = ".not_terminated";
	
	/* get the time, convert to broken-down format and format */
	time(&temp_time);
	binary_time = localtime(&temp_time);

	if (opt == 0)
	{
		strftime(start_time, timesize, fmt, binary_time); /* store for later */
		/* build the actual filename */
		strncpy(trail_name, start_time, timesize);
		strncat(trail_name, not_term, sizeof(not_term));
	}
	else if (opt == 1)
	{
		/* first, back up the filename for rename() */
		strncpy(trail_name_old, trail_name, strlen(trail_name));
		/* now, make the new name */
		strncpy(trail_name, start_time, timesize);		/* copy start time */
		strftime(start_time, timesize, fmt, binary_time); /* get end time */
		strncat(trail_name, period, 1);
		strncat(trail_name, start_time, timesize);		/* put them together */	
	}
	else
		error("set_filename called wrong");
		
	return;
}

/* spit an error and bail */
void error(char *s)
{
	printf("%s\n", s);
	exit(1);
}
/* uses perror to get the errno info */
void error_sys(char *s)
{
	perror(s);
	exit(1);	
}

/* apply the read in preferences here and send them to the driver */
void apply_prefs(void)
{
	int temp_size = user_prefs->bufsize; // save for comparison
	
	/* If we only have default prefs */
	if(user_prefs->default_flag == 1)
	{
		if(ioctl(device_fd, BSM_DEFAULT, 0))
			error_sys("auditd: can't set up with default prefs\n");
	}

	/* Set the memory area configuration - either default size or custom */
	if(ioctl(device_fd, BSM_CONFIG_BUF, user_prefs->bufsize))
		error_sys("auditd: can't set up with custom buffer size\n");
	/* since the size requested may be invalid (and so the default will be
	   used), need to find out the actual value */	
	user_prefs->bufsize = ioctl(device_fd, BSM_BUFSIZE, 0);
	printf("auditd: set buffers to %d bytes\n", user_prefs->bufsize);
	if (temp_size != user_prefs->bufsize)
	{
		printf("auditd: this isn't what you entered - seems you gave a ");
		printf("funny/invalid size. The driver doesn't like that...\n");
	}


	if(user_prefs->auditing_pid == 1)
    {
		if(ioctl(device_fd, BSM_CONFIG_PID, user_prefs->pid))
			error_sys("auditd: can't audit the given pid\n");

		printf("auditd: auditing pid %d only\n", user_prefs->pid);
    }

	if(user_prefs->auditing_gid == 1)
	{
		if(ioctl(device_fd, BSM_CONFIG_GID, user_prefs->gid))
			error_sys("auditd: can't audit the given gid\n");
			
		printf("auditd: auditing gid %d only\n", user_prefs->gid);
    }

	if(user_prefs->auditing_uid == 1)
	{
		if(ioctl(device_fd, BSM_CONFIG_UID, user_prefs->uid))
			error_sys("auditd: can't audit the given uid\n");
			
		printf("auditd: auditing uid %d only\n", user_prefs->uid);
    }


}

/* Process signals sent by the auditor driver */
void sig_handler(int sig_id)
{
	int events_lost; /* # of events lost */
	char* temp;
	
	switch(sig_id)
    {
	case SIGIO:
		if (busy == 0)
		{
			/* become busy, exchange buffers and write full one to trail */
			busy = 1;
	    
			temp  = buffer1;
			buffer1 = buffer2;
			buffer2 = temp;
		  
			if(write(trail_fd, buffer2, user_prefs->bufsize) < 0)
				error_sys("auditd: can't read from driver buffer, exiting");

			busy = 0;
			/* let the driver know */
			ioctl(device_fd, BSM_WRITTEN, 0);
			
			return;
		}
		/* else called with busy = 1, this is bad - cascade down */	
	default: 
		printf("auditd: signal problem, exiting\n");
    case SIGTERM:
    case SIGALRM:
		/* stop the driver and see if events were lost */
		ioctl(device_fd, BSM_STOP, 0);

		events_lost = ioctl(device_fd, BSM_LOST, 0);
		if(events_lost > 0)
		{
			printf("auditd: lost %d events, this is not good\n", events_lost);
			printf("auditd: perhaps you need a bigger buffer size\n");
		}
	}

	/* Write the current buffer out and quit */
	write(trail_fd, buffer1, user_prefs->bufsize);
	close(device_fd);
	close(trail_fd);
	/* Change the filename to start.end timestamp pair */
	set_filename(1);
	/* now have old and new filenames, rename */
	rename(trail_name_old, trail_name);
	printf("auditd: finished\n");
	exit(0);
}


/* Set everything up, then wait for data to write out */
int main(int argc, char* argv[])
{
	struct sigaction act; /* For sigaction() */
	int pid; /* Daemon pid */

	/* initial junk */
	user_prefs = new_prefs();
	busy = 0;
	
	printf("auditd: version %d.%d%d starting up\n", BSM_MAJOR_VERSION,
			BSM_MINOR_VERSION, BSM_PATCHLEVEL);
	
	/* handle command-line arguments and setup */
	set_prefs(argc, argv, user_prefs);
	
	/* become daemon, die if it fails */
	if((pid = fork()) < 0)
		error_sys("auditd: can't become daemon, exiting \n");
	else if(pid) /* I'm the parent, suicide */
		exit(0);
		
	printf("auditd: set up as daemon...\n");
	
	/* start a new session and set up signal handling */
	process_id = setsid();
	  
	act.sa_handler = sig_handler;
	sigfillset(&(act.sa_mask));	/* block all signals while handling */
	act.sa_flags = 0;			/* no flags */
	if(sigaction(SIGIO, &act, NULL))
		error_sys("auditd: problem setting signals, exiting");
	if(sigaction(SIGTERM, &act, NULL))
		error_sys("auditd: problem setting signals, exiting");
	if(sigaction(SIGALRM, &act, NULL))
		error_sys("auditd: problem setting signals, exiting");

	/* build the audit trail filename and open the trail up */	
	set_filename(0);

	if((trail_fd = open(trail_name,
  		O_CREAT | O_RDWR | O_TRUNC, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH)) < 0)
  		error_sys("auditd: can't open audit trail, exiting");

	printf("auditd: audit trail open, file name is %s \n", trail_name);

	if((device_fd = open(device_name, O_RDONLY, 0)) < 0)
		error_sys("auditd: can't open driver");

	printf("auditd: driver open, setting prefs\n");
	apply_prefs();
	printf("auditd: mapping driver memory\n");
	
	/* map the driver's buffers (in the kernel) to our address space */
	/* bufsize << 1 because we have two of them in one contiguous chunk */
	buffer1 = mmap(NULL, user_prefs->bufsize << 1, PROT_READ | PROT_WRITE, MAP_PRIVATE,
				device_fd, 0);
	if(buffer1 == MAP_FAILED) /* close things up, delete the empty file, and die */
	{
		close(trail_fd);
		close(device_fd);
		unlink(trail_name);
		error_sys("auditd: memory mapping failed, exiting");
    }
	
	/* Set up secondary buffer */
	buffer2 = buffer1 + user_prefs->bufsize;

	printf("auditd: done, starting audit\n");

	if(ioctl(device_fd, BSM_START, 0) < 0)
	{
		close(trail_fd);
		close(device_fd);
		unlink(trail_name);
		error_sys("auditd: can't start auditing, exiting");
	}

	/* sleep until we're signaled to get some data by the auditor */
	while(1)
		pause();

	/* if we're here, something went very wrong */
	error_sys("oops");
}
